﻿#define _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <direct.h>

/* --- プログラムの説明 --- */
char *filen = "P2-6osart_ptn.c";
char *title = "OSSART法で雑音パターンを変えて再構成するプログラム";
/*
  OS（Ordered Subset）を用いている
  雑音パターンを変えながら平均，RMSEを記録する
*/

/* 入力 */
char *menu[] = { // 入力の際のコメント（入力変数とリンク）
	" 1. 原画像のファイル名       ",
	" 2. 画像の幅                 ",
	" 3. 画像の高さ               ",
	" 4. 投影の幅                 ",
	" 5. 投影数                   ",
	" 6. 画素長 (cm/pixel)        ",
	" 7. TVの重み係数             ",
	" 8. 重み (0:なし，1:あり)    ",
	" 9. サブセットの数           ",
	"10. マスク画像のファイル名   ",
	"11. 投影の入射強度           ",
	"12. 雑音の種類 (0:ポアソン，1：ガウス) ",
	"13. 雑音レベル (dB)   (ガウス雑音のみ) ",
	"14. 雑音乱数のパターンの数   ",
	"15. 繰り返し回数             ",
};

/* 出力 */
/*
  再構成した画像データのファイル
  n***.img   : 更新された画像
  ※ "***"には繰り返しの回数が挿入される（例：001）
  _mean_**.csv : 再構成画像の平均値
  _rmse_**.csv : 再構成画像のRMSE
  ※ "**"には0が全体，1～がROI番号が挿入される
  ※ 自動で作成されたフォルダに格納される
*/

#define  PI  3.14159265358979

// プロトタイプ宣言
void read_data(char *fi, double *img, int size);
void write_data(char *fi, double *img, int size);
void write_rmse_id(double *mean, double *rmse, int ma, int nit, int ptn, int id);
void detection_probability(int *cx, double *cc, int px, int pa, int p2, int nx, int ny, int ns);
void projection_c(double *prj, int px, int pa, double *img, int nx, int ny, int ns, double pl);
void noise_xct_poisson(double *prj2, double *prj, int px, int pa, int i0, int id);
void noise_xct_gauss(double *prj2, double *prj, int px, int pa, int i0, double db, int id);
void recon_ossart_id(int id);

// グローバル変数の宣言と初期値設定
char    g_f1[50] = "xdisk_6.img";   // 原画像のファイル名
char    g_f2[50] = "xdisk_6_mask.img"; // マスク画像のファイル名
char    g_fd[256] = "";             // 保存用のフォルダ名

int     g_px = 256;            // 投影の幅
int     g_pa = 256;            // 投影数
int     g_nx = 256;            // 画像の幅
int     g_ny = 256;            // 画像の高さ
double  g_pl = 0.078125;       // 画素長 (cm/pixel)
double  g_beta2 = 0.00002;     // TVの重み係数
double  g_wt = 1;              // 重み (0:なし，1:あり) 
int     g_ss = 32;             // サブセットの数
int     g_I0 = 2000;           // X線の初期強度（ブランクのカウント数）
int     g_nt = 0;              // 雑音の種類 (0:ポアソン，1：ガウス)
double  g_db = 45;             // 雑音レベル（dB）ガウス雑音のみ
int     g_ptn = 10;            // 乱数のパターン数（idum=-1～-g_ptn）
int     g_nit = 20;           // 繰り返し回数
int     g_p2 = 1;              // 投影の全角度 [2]π or [1]π
int     g_ns = 3;              // Cijの検出器分配数
int     g_ma = 0;              // マスク（ROI）の数+1（マスク画像のラベル数+1）

double *g_img;                 // 原画像データ領域
double *g_pr0;                 // 雑音なし投影データ領域
double *g_prj;                 // 雑音あり投影データ領域
int    *g_cx;                  // 検出確率の中央のx座標
double *g_cc;                  // 検出確率Cijの値
double *g_msk;                 // マスク画像（1，2，3，・・・）
double *g_rmse;                // RMSEのストック配列
double *g_mean;                // 平均のストック配列

void getparameter()
{
	int   i= 0;
	char  dat[256];

	// 変数への値の入力
	fprintf( stdout, "\n%s\n\n", title);

	fprintf( stdout, " %s [%s] :", menu[i++], g_f1 );
	if (*fgets(dat, 256, stdin) != '\n') { dat[strlen(dat) - 1] = '\0'; strcpy(g_f1, dat); }
	fprintf(stdout, " %s [%d] :", menu[i++], g_nx);
	if (*fgets(dat, 256, stdin) != '\n')  g_nx = atoi(dat);
	fprintf(stdout, " %s [%d] :", menu[i++], g_ny);
	if (*fgets(dat, 256, stdin) != '\n')  g_ny = atoi(dat);
	fprintf( stdout, " %s [%d] :", menu[i++], g_px );
	if (*fgets(dat, 256, stdin) != '\n')  g_px = atoi(dat);
	fprintf( stdout, " %s [%d] :", menu[i++], g_pa );
	if (*fgets(dat, 256, stdin) != '\n')  g_pa = atoi(dat);
	fprintf(stdout, " %s [%f] :", menu[i++], g_pl);
	if (*fgets(dat, 256, stdin) != '\n')  g_pl = atof(dat);
	fprintf(stdout, " %s [%f] :", menu[i++], g_beta2);
	if (*fgets(dat, 256, stdin) != '\n')  g_beta2 = atof(dat);
	fprintf(stdout, " %s [%f] :", menu[i++], g_wt);
	if (*fgets(dat, 256, stdin) != '\n')  g_wt = atof(dat);
	fprintf(stdout, " %s [%d] :", menu[i++], g_ss);
	if (*fgets(dat, 256, stdin) != '\n')  g_ss = atoi(dat);
	fprintf(stdout, " %s [%s] :", menu[i++], g_f2);
	if (*fgets(dat, 256, stdin) != '\n') { dat[strlen(dat) - 1] = '\0'; strcpy(g_f2, dat); }
	fprintf(stdout, " %s [%d] :", menu[i++], g_I0);
	if (*fgets(dat, 256, stdin) != '\n')  g_I0 = atoi(dat);
	fprintf(stdout, " %s [%d] :", menu[i++], g_nt);
	if (*fgets(dat, 256, stdin) != '\n')  g_nt = atoi(dat);
	fprintf(stdout, " %s [%f] :", menu[i++], g_db);
	if (*fgets(dat, 256, stdin) != '\n')  g_db = atof(dat);
	fprintf(stdout, " %s [%d] :", menu[i++], g_ptn);
	if (*fgets(dat, 256, stdin) != '\n')  g_ptn = atoi(dat);
	fprintf(stdout, " %s [%d] :", menu[i++], g_nit);
	if (*fgets(dat, 256, stdin) != '\n')  g_nit = atoi(dat);

}

int main(void)
{
	int i;
	char fi[256];
	time_t  timer;
	struct tm *local;

	// プログラムで使用する変数の入力
	getparameter();

	// メモリを動的に確保
	g_img = (double *)malloc((size_t)g_nx*g_ny * sizeof(double));
	g_pr0 = (double *)malloc((size_t)g_px*g_pa * sizeof(double));
	g_prj = (double *)malloc((size_t)g_px*g_pa * sizeof(double));
	g_msk = (double *)malloc((size_t)g_nx*g_ny * sizeof(double));
	g_cx = (int *)malloc((size_t)g_pa*g_nx*g_ny * sizeof(int));
	g_cc = (double *)malloc((size_t)g_pa*g_nx*g_ny*g_ns * sizeof(double));

	// 出力フォルダの作成
	timer = time(NULL);
	local = localtime(&timer); // 地方時に変換

	sprintf(g_fd, "%d%02d%02d_%02d%02d%02d_%s_%s_TV%f_W%.1f_I0%d",
		local->tm_year + 1900, local->tm_mon + 1, local->tm_mday,
		local->tm_hour, local->tm_min, local->tm_sec,
		filen, g_f1, g_beta2, g_wt, g_I0);
	_mkdir(g_fd);

	// 検出確率の計算
	detection_probability(g_cx, g_cc, g_px, g_pa, g_p2, g_nx, g_ny, g_ns);

	printf(" *** Read projection data   ***\n");
	// オリジナル画像とマスク画像の入力
	read_data(g_f1, g_img, g_nx*g_ny);
	read_data(g_f2, g_msk, g_nx*g_ny);
	
	// マスク画像のラベル数の算出
	for (i = 0; i < g_nx*g_ny; i++)
	{
		if (g_ma < (int)g_msk[i]) g_ma = (int)g_msk[i];
	}
	g_ma++; // 全体のRMSE算出を数に加える

	// ラベル数に合わせてメモリを動的に確保
	g_mean = (double *)malloc((size_t)(g_nit + 1)*g_ptn*g_ma * sizeof(double));
	g_rmse = (double *)malloc((size_t)(g_nit + 1)*g_ptn*g_ma * sizeof(double));

	// 雑音なし投影の作成
	projection_c(g_pr0, g_px, g_pa, g_img, g_nx, g_ny, g_ns, g_pl);
	sprintf(fi, "%s/_org.prj", g_fd);
	write_data(fi, g_pr0, g_px*g_pa);

	printf(" *** OSAART reconstruction ***\n");
	// OSAART法の実行（雑音パターンを変えて実行）
	for (i = 0; i < g_ptn; i++)
	{
		// 雑音入りの投影作成
		if (g_nt == 0) noise_xct_poisson(g_prj, g_pr0, g_px, g_pa, g_I0, i + 1);
		else           noise_xct_gauss(g_prj, g_pr0, g_px, g_pa, g_I0, g_db, i + 1);
		sprintf(fi, "%s/id%d__org.prj", g_fd, i + 1);
		write_data(fi, g_prj, g_px*g_pa);

		// OSSART法での再構成（パターンあり）
		recon_ossart_id(i);

		// RMSEの出力
		write_rmse_id(g_mean, g_rmse, g_ma, g_nit, g_ptn, i);
	}

	free(g_img);
	free(g_prj);
	free(g_pr0);
	free(g_msk);
	free(g_cx);
	free(g_cc);
	free(g_mean);
	free(g_rmse);
	return 0;
}

// *** 画像領域の初期化 ***
// double *img;  // 画像領域
// int    size;  // 画像領域のデータ数（画素数）
// double  val;  // 初期化する値
void init(double *img, int size, double val)
{
	int  i;
	for (i = 0; i < size; i++)
		img[i] = val;
}

// *** 2次元画像データの入力（float型ファイルをdoubleとして入力）***
// char *fi;    // 入力画像のファイル名
// double *img; // 入力画像データ
// int  size;   // 入力画像のサイズ（幅×高さ pixel）
void read_data(char *fi, double *img, int size)
{
	int i;
	float buff;
	FILE   *fp;

	if ((fp = fopen(fi, "rb")) == NULL) {
		fprintf(stderr, " エラー：ファイルが開きません[%s].\n", fi);
		exit(1);
	}
	for (i = 0; i < size; i++)
	{
		fread(&buff, sizeof(float), 1, fp);
		img[i] = buff;
	}
	fclose(fp);
}

// *** 2次元画像データの出力（doubleデータをfloat型として出力）***
// char *fi;    // 出力画像のファイル名
// double *img; // 出力画像データ
// int  size;   // 出力画像のサイズ（幅×高さ pixel）
void write_data(char *fi, double *img, int size)
{
	int i;
	float buff;
	FILE   *fp;

	if ((fp = fopen(fi, "wb")) == NULL) {
		fprintf(stderr, " エラー：ファイルが開きません[%s].\n", fi);
		exit(1);
	}
	for (i = 0; i < size; i++)
	{
		buff = (float)img[i];
		fwrite(&buff, sizeof(float), 1, fp);
	}
	fclose(fp);
}

// *** 平均とRMSEをテキストファイルに書き込む関数 ***
// double *mean; // 平均値のデータ
// double *rmse; // RMSEのデータ
// int    ma;    // マスク（ROI）の数
// int    nit;   // 繰り返し回数
// int    ptn;   // 雑音の全パターン数
// int    id;    // 雑音パターンの番号
void write_rmse_id(double *mean, double *rmse, int ma, int nit, int ptn, int id)
{
	char    fi[256];
	FILE    *fp;
	int     i, j, k;

	for (i = 0; i < ma; i++)
	{
		// 平均のファイル出力
		sprintf(fi, "%s/_mean_%d.csv", g_fd, i);
		if (id == 0)
		{
			if ((fp = fopen(fi, "w")) == NULL)
			{
				fprintf(stderr, " Error : file open [%s].\n", fi);
				exit(1);
			}
			fprintf(fp, "%s\n", fi);
			fprintf(fp, "No.");
			for (j = 0; j <= nit; j++)
				fprintf(fp, ",%d", j);
			fprintf(fp, "\n");
			fclose(fp);
		}
		if ((fp = fopen(fi, "a")) == NULL)
		{
			fprintf(stderr, " Error : file open [%s].\n", fi);
			exit(1);
		}
		fprintf(fp, "%d", id + 1);
		for (j = 0; j <= nit; j++)
			fprintf(fp, ",%f", mean[i*g_ptn*(nit + 1) + id*(nit + 1) + j]);
		fprintf(fp, "\n");

		// 繰り返し回数ごとの平均
		if (id + 1 == g_ptn)
		{
			fprintf(fp, "Ave.");
			for (j = 0; j <= nit; j++)
			{
				double ave = 0;
				for (k = 0; k <= id; k++)
					ave += mean[i*g_ptn*(nit + 1) + k*(nit + 1) + j];
				ave /= g_ptn;
				fprintf(fp, ",%f", ave);
			}
			fprintf(fp, "\n");
		}
		fclose(fp);

		// RMSEのファイル出力
		sprintf(fi, "%s/_rmse_%d.csv", g_fd, i);
		if (id == 0)
		{
			if ((fp = fopen(fi, "w")) == NULL)
			{
				fprintf(stderr, " Error : file open [%s].\n", fi);
				exit(1);
			}
			fprintf(fp, "%s\n", fi);
			fprintf(fp, "No.");
			for (j = 0; j <= nit; j++)
				fprintf(fp, ",%d", j);
			fprintf(fp, "\n");
			fclose(fp);
		}
		if ((fp = fopen(fi, "a")) == NULL)
		{
			fprintf(stderr, " Error : file open [%s].\n", fi);
			exit(1);
		}
		fprintf(fp, "%d", id + 1);
		for (j = 0; j <= nit; j++)
			fprintf(fp, ",%f", rmse[i*g_ptn*(nit + 1) + id*(nit + 1) + j]);
		fprintf(fp, "\n");

		// 繰り返し回数ごとの平均
		if (id + 1 == g_ptn)
		{
			fprintf(fp, "Ave.");
			for (j = 0; j <= nit; j++)
			{
				double ave = 0;
				for (k = 0; k <= id; k++)
					ave += rmse[i*g_ptn*(nit + 1) + k*(nit + 1) + j];
				ave /= g_ptn;
				fprintf(fp, ",%f", ave);
			}
			fprintf(fp, "\n");
		}
		fclose(fp);
	}
}

// *** 平均とRMSEを事前に計算する関数 ***
// double *img; // 対象の画像
// double *org; // 原画像
// double *msk; // マスク画像
// int     nx;  // 画像の幅
// int     ny;  // 画像の高さ
// int     ma;  // マスク（ROI）の数
// int     id;  // 雑音パターン番号
// int     itn; // 繰り返し番号
void make_rmse_id(double *img, double *org, double *msk, int nx, int ny, int ma, int id, int itn)
{
	int  i, j, count;
	double m1 = 0.0;
	double v1 = 0.0;
	double v2 = 0.0;

	// 全体の平均とRMSE計算
	for (j = 0; j < nx*ny; j++)
	{
		m1 += img[j];
		v1 += (img[j] - org[j])*(img[j] - org[j]);
		v2 += org[j] * org[j];
	}
	m1 /= (g_nx*g_ny);
	v1 = 100.*sqrt(v1 / v2);

	g_mean[id*(g_nit + 1) + itn] = (float)m1;
	g_rmse[id*(g_nit + 1) + itn] = (float)v1;

	// マスク（ROI）の平均とRMSE計算
	for (i = 1; i < ma; i++)
	{
		count = 0;
		m1 = v1 = v2 = 0;
		for (j = 0; j < nx*ny; j++)
		{
			if (msk[j] == (float)i)
			{
				m1 += img[j];
				v1 += (img[j] - org[j])*(img[j] - org[j]);
				v2 += org[j] * org[j];
				count++;
			}
		}
		m1 /= count;
		v1 = 100.*sqrt(v1 / v2);

		g_mean[i*g_ptn*(g_nit + 1) + id*(g_nit + 1) + itn] = (float)m1;
		g_rmse[i*g_ptn*(g_nit + 1) + id*(g_nit + 1) + itn] = (float)v1;
	}
}

// *** RMSEの計算結果を返す ***
// double *im1; // 評価用画像データ
// double *im0; // 原画像データ
// int  size;   // 画像のサイズ（幅×高さ pixel）
double calc_rmse(double *im1, double *im0, int size)
{
	int     i;
	double  sum = 0, sum2 = 0;

	for (i = 0; i < size; i++)
	{
		sum += (im1[i] - im0[i])*(im1[i] - im0[i]);
		sum2 += im0[i] * im0[i];
	}

	return 100.*sqrt(sum) / sqrt(sum2);
}

// *** X線CTの投影データにポアソン雑音を付加する関数 ***
// double  *prj2; // 雑音を付加した投影データ
// double  *prj;  // 雑音なしの投影データ
// int     px;    // 投影の幅
// int     pa;    // 投影数
// double  i0;    // X線の初期強度
// int     id;    // ランダムパターン番号
void noise_xct_poisson(double *prj2, double *prj, int px, int pa, int i0, int id)
{
	int     i;
	long    idum = (long)(-id);
	double	ixt, ixt0;
	double  poidev(double, long *);

	for (i = 0; i < px*pa; i++)
	{
		ixt0 = i0*exp(-prj[i]);     // 計測データに変換
		ixt = poidev(ixt0, &idum);  // ノイズを挿入
		if (ixt < 1.)	ixt = 1;  // カウントが0だったら1にする

		prj2[i] = log(i0 / ixt);  // 投影データに変換
	}
}

// *** X線CTの投影データへガウス雑音を付加する関数 ***
// double  *prj2; // 雑音を付加した投影データ
// double  *prj;  // 雑音なしの投影データ
// int     px;    // 投影の幅
// int     pa;    // 投影数
// double  i0;    // X線の初期強度
// double  db;    // 雑音レベル（dB）
// int     id;    // ランダムパターン番号
void noise_xct_gauss(double *prj2, double *prj, int px, int pa, int i0, double db, int id)
{
	int     i;
	long    idum = (long)(-id);
	double	ixt, sum, ave, sd;
	double  gasdev(long[]);

	// 計測データへの変換と平均値の計算
	sum = 0.0;
	for (i = 0; i < px*pa; i++)
	{
		ave = i0*exp(-prj[i]);
		sum += ave;
		prj2[i] = ave;
	}
	ave = sum / (px*pa);

	// 雑音の標準偏差
	sd = ave * pow(10.0, -db / 20.0);

	// 雑音の付加
	for (i = 0; i < px*pa; i++)
	{
		ixt = prj2[i] + sd*gasdev(&idum);
		if (ixt < 1.)	ixt = 1;  // カウントが0だったら1にする

		prj2[i] = log(i0 / ixt);  // 投影データに変換
	}
}

// *** ガウス型のlow-passフィルタ ***
// double  *prw; // 処理後の投影データ
// double  *prj; // 処理前の投影データ
// int      px;  // 投影の幅
// int      pa;  // 投影数
void lowpass_filter(double *prw, double *prj, int px, int pa)
{
	int i, j, k, n;
	int c = 4;
	double *h;
	double fwhm = 1;
	double s = fwhm / (2 * sqrt(2 * log(2.0)));

	h = (double *)malloc((size_t)c*sizeof(double));

	// ガウシアンフィルタ
	for (n = 0; n < c; n++)
	{
		h[n] = 1 / (sqrt(2 * PI)*s)*exp(-n*n / (2 * s*s));
	}

	// 重畳積分によるフィルタリング
	for (i = 0; i < pa; i++)
	{
		for (j = 0; j < px; j++)
		{
			prw[i*px + j] = 0;
			for (k = 0; k < px; k++)
			{
				// 重畳積分
				if (abs(j - k) < c)
					prw[i*px + j] += prj[i*px + k] * h[abs(j - k)];
			}
		}
	}

	free(h);
}

// *** 検出確率のマトリクスを作成する関数 ***
// int    *cx;  // 検出確率の検出位置xの値
// double *cc;  // 検出確率の値
// int    px;   // 投影の動径方向の数
// int    pa;   // 投影の角度方向の数
// int    p2;   // 投影データの全角度（[2]πか[1]π）
// int    nx;   // 画像のx方向の数
// int    ny;   // 画像のy方向の数
// int    ns;   // 1画素から入る検出器の幅
void detection_probability(int *cx, double *cc, int px, int pa, int p2, int nx, int ny, int ns)
{
	int     i, j, k, ix;
	double  x, y, xx, th, a, b, x05, d, si, co;
	double  cca[3];

	for (i = 0; i < pa*nx*ny; i++)
		cx[i] = 0;
	for (i = 0; i < pa*nx*ny*ns; i++)
		cc[i] = 0;

	for (k = 0; k < pa; k++)
	{
		fprintf(stderr, "\r *** Make cij [%3d/%3d]", k, pa);
		th = p2 * PI * k / pa;
		si = sin(th);
		co = cos(th);
		if (fabs(si) > fabs(co))
		{
			a = fabs(si);
			b = fabs(co);
		}
		else
		{
			a = fabs(co);
			b = fabs(si);
		}
		for (i = 0; i < ny; i++)
		{
			y = ny / 2 - i;
			for (j = 0; j < nx; j++)
			{
				x = j - nx / 2;
				xx = x * co + y * si;

				cca[0] = cca[1] = cca[2] = 0.0;

				ix = (int)(floor(xx + 0.5));
				if (ix + px / 2 < 1 || ix + px / 2 > px - 2) continue;

				x05 = ix - 0.5;
				if ((d = x05 - (xx - (a - b) / 2)) > 0.0)
					cca[0] = b / (2 * a) + d / a;
				else if ((d = x05 - (xx - (a + b) / 2)) > 0.0)
					cca[0] = d * d / (2 * a * b);

				x05 = ix + 0.5;
				if ((d = xx + (a - b) / 2 - x05) > 0.0)
					cca[2] = b / (2 * a) + d / a;
				else if ((d = xx + (a + b) / 2 - x05) > 0.0)
					cca[2] = d * d / (2 * a * b);

				cca[1] = 1.0 - cca[0] - cca[2];

				cx[k*nx*ny + i * nx + j] = ix + px / 2 - ns / 2;
				cc[(k*nx*ny + i * nx + j)*ns + 0] = cca[0];
				cc[(k*nx*ny + i * nx + j)*ns + 1] = cca[1];
				cc[(k*nx*ny + i * nx + j)*ns + 2] = cca[2];
			}
		}
	}
	fprintf(stderr, "\r *** Make cij [%3d/%3d]\n", k, pa);
}

// *** 検出確率を使って投影データ1つを作成する関数 ***
// double  *prj; // 作成する投影データ
// int     px;   // 投影の幅
// int     pa;   // 投影数
// double  *img; // もとになる画像データ
// int     nx;   // 画像の幅
// int     ny;   // 画像の高さ
// int     ns;   // Cijの検出器分配数
// double  pl;   // 画素長
// int     k;    // 角度方向の番号
void projection_c_one(double *prj, int px, int pa, double *img, int nx, int ny, int ns, double pl, int k)
{
	int    i, j;

	for (i = 0; i < px; i++)
		prj[k*px+i] = 0;

	for (i = 0; i < nx*ny; i++)
	{
		for (j = 0; j < ns; j++)
		{
			int jj = g_cx[k*nx*ny + i] + j;
			if (jj < 0 || jj > px - 1) continue;
			prj[k*px + jj] += (g_cc[(k*nx*ny + i)*ns + j] * img[i] * pl);
		}
	}
}

// *** 検出確率を使って投影データを作成する関数 ***
// double  *prj; // 作成する投影データ
// int     px;   // 投影の幅
// int     pa;   // 投影数
// double  *img; // もとになる画像データ
// int     nx;   // 画像の幅
// int     ny;   // 画像の高さ
// int     ns;   // Cijの検出器分配数
// double  pl;   // 画素長
void projection_c(double *prj, int px, int pa, double *img, int nx, int ny, int ns, double pl)
{
	int    i, j, k;

	for (i = 0; i < px*pa; i++)
		prj[i] = 0;

	for (k = 0; k < pa; k++)
	{
		for (i = 0; i < nx*ny; i++)
		{
			for (j = 0; j < ns; j++)
			{
				int jj = g_cx[k*nx*ny + i] + j;
				if (jj < 0 || jj > px - 1) continue;
				prj[k*px + jj] += (g_cc[(k*nx*ny + i)*ns + j] * img[i] * pl);
			}
		}
	}
}

// *** 検出確率を使って投影データ1つから逆投影する関数 ***
// double  *img; // 作成する画像データ
// int     nx;   // 画像の幅
// int     ny;   // 画像の高さ
// double  *prj; // もとになる投影データ
// int     px;   // 投影の幅
// int     pa;   // 投影数
// int     ns;   // Cijの検出器分配数
// double  pl;   // 画素長
// int     k;    // 角度方向の番号
void backprojection_c_one(double *img, int nx, int ny, double *prj, int px, int pa, int ns, double pl, int k)
{
	int    i, j;

	for (i = 0; i < nx*ny; i++)
	{
		for (j = 0; j < ns; j++)
		{
			int jj = g_cx[k*nx*ny + i] + j;
			if (jj < 0 || jj > px - 1) continue;
			img[i] += (g_cc[(k*nx*ny + i)*ns + j] * prj[k*px + jj] / pl);
		}
	}
}

// *** ∇TV（Total Variationの勾配）の計算 ***
// double  *ntv;  // ∇TVの計算結果
// double  *img;  // もとになる画像データ
// int     nx;    // 画像の幅
// int     ny;    // 画像の高さ
void nabla_tv(double *ntv, double *img, int nx, int ny)
{
	int     i, j, k, x[3], y[3];
	double  fil[9], tv1, tv2, ep = 0.0001;
	double  beta = 1.0;

	// TVの計算
	for (i = 0; i < ny; i++)
	{
		y[0] = (i + ny - 1) % ny;
		y[1] = i;
		y[2] = (i + 1) % ny;
		for (j = 0; j < nx; j++)
		{
			x[0] = (j + nx - 1) % nx;
			x[1] = j;
			x[2] = (j + 1) % nx;
			for (k = 0; k < 9; k++)
				fil[k] = img[y[k / 3] * nx + x[k % 3]];
			tv1 = (fil[4] - fil[3]) / sqrt((fil[4] - fil[3])*(fil[4] - fil[3]) + (fil[6] - fil[3])*(fil[6] - fil[3]) + ep * ep)
				+ (fil[4] - fil[1]) / sqrt((fil[2] - fil[1])*(fil[2] - fil[1]) + (fil[4] - fil[1])*(fil[4] - fil[1]) + ep * ep)
				- (fil[5] + fil[7] - 2 * fil[4]) / sqrt((fil[5] - fil[4])*(fil[5] - fil[4]) + (fil[7] - fil[4])*(fil[7] - fil[4]) + ep * ep);
			tv2 = (fil[4] - fil[1]) / sqrt((fil[4] - fil[1])*(fil[4] - fil[1]) + (fil[0] - fil[1])*(fil[0] - fil[1]) + ep * ep)
				+ (fil[4] - fil[5]) / sqrt((fil[8] - fil[5])*(fil[8] - fil[5]) + (fil[4] - fil[5])*(fil[4] - fil[5]) + ep * ep)
				- (fil[7] + fil[3] - 2 * fil[4]) / sqrt((fil[7] - fil[4])*(fil[7] - fil[4]) + (fil[3] - fil[4])*(fil[3] - fil[4]) + ep * ep);

			ntv[i*nx + j] = beta * (tv1 + tv2) / 2;
		}
	}
}

// *****************************************
// ***  逐次近似投影再構成 (OS-SART法)   ***
// *****************************************
void recon_ossart_id(int id)
{
	int     i, j, k, jj, m1, m2;
	char    fi[256];
	int     *sub;   // サブセットの順番
	double  *im1;   // 仮定画像のデータ領域
	double  *prw;   // 重み付け最小二乗法用のデータ領域
	double  *aprj;  // 仮定画像からの投影
	double  *rprj;  // 実投影との差（重み付けあり）
	double  *aimg;  // 差の逆投影画像のデータ領域
	double  *bimg;  // 正則化項のデータ領域

	// メモリを動的に確保
	prw = (double *)malloc((size_t)g_px*g_pa * sizeof(double));
	im1 = (double *)malloc((size_t)g_nx*g_ny * sizeof(double));
	aprj = (double *)malloc((size_t)g_px*g_pa * sizeof(double));
	rprj = (double *)malloc((size_t)g_px*g_pa * sizeof(double));
	aimg = (double *)malloc((size_t)g_nx*g_ny * sizeof(double));
	bimg = (double *)malloc((size_t)g_nx*g_ny * sizeof(double));

	// サブセットの順番を決定する
	sub = (int *)malloc(g_ss * sizeof(int));
	k = 0;
	for (i = 0; i < 32; i++)
		k += (g_ss >> i) & 1;
	if (k == 1)
	{
		m1 = 0;
		sub[m1++] = 0;
		for (i = g_ss, m2 = 1; i > 1; i /= 2, m2 *= 2)
		{
			for (j = 0; j < m2; j++)
				sub[m1++] = sub[j] + i / 2;
		}
	}
	else
	{
		for (i = 0; i < g_ss; i++)
			sub[i] = i;
	}

	// 重み付け用の投影作成（フィルタリングによる平滑化）
	if (g_wt <= 0.0)
	{
		init(prw, g_px*g_pa, 0.0);
	}
	else
	{
		lowpass_filter(prw, g_prj, g_px, g_pa);
	}

	// OSAART itaration
	// ① 初期画像を仮定する
	init(im1, g_nx*g_ny, 0.1);

	// RMSEの出力
	make_rmse_id(im1, g_img, g_msk, g_nx, g_ny, g_ma, id, 0);

	// 逐次近似の繰り返し
	for (i = 0; i < g_nit; i++)
	{
		fprintf(stderr, "\r *** OSAART iteration id_%02d [%2d/%2d]", id, i + 1, g_nit);

		for (k = 0; k < g_ss; k++)
		{
			// ② 仮定画像から投影を計算する
			for (j = sub[k]; j < g_pa; j += g_ss)
				projection_c_one(aprj, g_px, g_pa, im1, g_nx, g_ny, g_ns, g_pl, j);

			// ③ 投影データと，②で計算した投影との差を計算する
			// 投影データの分散(prw)を組み入れている
			for (j = sub[k]; j < g_pa; j += g_ss)
			{
				for (jj = 0; jj < g_px; jj++)
				{
					rprj[j*g_px + jj] = (aprj[j*g_px + jj] - g_prj[j*g_px + jj])*exp(-prw[j*g_px + jj]);
				}
			}

			// ④ ③で計算された差を逆投影する
			for (j = 0; j < g_nx*g_ny; j++)
				aimg[j] = 0;
			for (j = sub[k]; j < g_pa; j += g_ss)
				backprojection_c_one(aimg, g_nx, g_ny, rprj, g_px, g_pa, g_ns, g_pl, j);
			for (j = 0; j < g_nx*g_ny; j++)
				aimg[j] /= g_px*g_pa / g_ss;

			// ⑤ 逆投影画像を仮定画像に重み付けして加える（画像の更新）
			for (j = 0; j < g_nx*g_ny; j++)
			{
				im1[j] -= aimg[j];

				if (g_beta2 == 0)
					if (im1[j] < 0.0) im1[j] = 0;
			}

			// ⑥ *** TV補正（繰り返しを用いる）
			if (g_beta2 != 0)
			{
				int     tvit = 1; // tvの繰り返し回数

				for (j = 0; j < g_nx*g_ny; j++)
					bimg[j] = 0;

				for (jj = 0; jj < tvit; jj++)
				{
					// TV用の補正係数の作成
					nabla_tv(bimg, im1, g_nx, g_ny);

					// TVの補正
					for (j = 0; j < g_nx*g_ny; j++)
					{
						im1[j] -= g_beta2*bimg[j];
						if (im1[j] < 0.0) im1[j] = 0;
					}
				}
			}
			// *** TV補正ここまで
		}

		// 結果画像の出力
		if (i < 10 || i % 10 == 9)
		{
			sprintf(fi, "%s\\id%d_n%03d.img", g_fd, id + 1, i + 1);
			write_data(fi, im1, g_nx*g_ny);
		}
		// RMSEのファイルへの出力
		make_rmse_id(im1, g_img, g_msk, g_nx, g_ny, g_ma, id, i + 1);
		fprintf(stderr, ", RMSE = %f\n", calc_rmse(im1, g_img, g_nx*g_ny));
	}
	printf("\n");

	free(prw);
	free(im1);
	free(aprj);
	free(rprj);
	free(aimg);
	free(bimg);
}

// **************************************
// 乱数発生用の関数
// (参考文献：NUMERICAL RECIPES in C)
// **************************************

#define  IA      16807
#define  IM      2147483647
#define  AM      (1.0/IM)
#define  IQ      127773
#define  IR      2836
#define  NTAB    32
#define  NDIV    (1+(IM-1)/NTAB)
#define  EPS     1.2e-7
#define  RNMX    (1.0-EPS)

// *** 一様乱数を発生させる関数 {0.0,1.0} ***
// long *idum;  // 初期化（数字を負値にして変えるとパターンが変わる）
double ran1(long idum[])
{
	int          j;
	long         k;
	static long  iy = 0;
	static long  iv[NTAB];
	double       temp;

	/* 初期化 */
	if (idum[0] <= 0 || !iy)
	{
		if (-(idum[0]) < 1)  idum[0] = 1;
		else                idum[0] = -idum[0];
		for (j = NTAB + 7; j >= 0; j--)
		{
			k = idum[0] / IQ;
			idum[0] = IA*(idum[0] - k*IQ) - IR*k;
			if (idum[0] < 0)  idum[0] += IM;
			if (j < NTAB)     iv[j] = idum[0];
		}
		iy = iv[0];
	}

	/* 乱数の作成 */
	k = idum[0] / IQ;
	idum[0] = IA*(idum[0] - k*IQ) - IR*k;
	if (idum[0] < 0)  idum[0] += IM;
	j = iy / NDIV;
	iy = iv[j];
	iv[j] = idum[0];
	if ((temp = AM*iy) > RNMX) return RNMX;
	else                    return temp;
}

// *** ガウス型のノイズを発生させる関数 ***
// long *idum;  // 一様乱数の初期化に用いる
// ガウス関数（平均0，標準偏差1）
double gasdev(long idum[])
{
	double         fac, rsq, v1, v2;
	static int     iset = 0;
	static double  gset = 0.0;

	if (iset == 0)
	{
		do
		{
			v1 = 2.0 * ran1(idum) - 1.0;
			v2 = 2.0 * ran1(idum) - 1.0;
			rsq = v1 * v1 + v2 * v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac = sqrt(-2.0 * log(rsq) / rsq);
		gset = v1 * fac;
		iset = 1;
		return v2 * fac;
	}
	else
	{
		iset = 0;
		return gset;
	}
}

// *** γ関数のln（自然対数）を返す関数 ***
// double  xx;  // 入力の値
// return value is ln gamma(xx) when xx > 0
double gammln(double xx)
{
	int           j;
	double        x, y, tmp, ser;
	static double cof[6] =
	{ 76.18009172947146,      -86.50532032941677,
		24.01409824083091,       -1.231739572450155,
		0.1208650973866179e-2,  -0.5395239384953e-5 };

	y = x = xx;
	tmp = x + 5.5;
	tmp -= (x + 0.5)*log(tmp);
	ser = 1.000000000190015;
	for (j = 0; j <= 5; j++)
		ser += cof[j] / ++y;
	return  (-tmp + log(2.5066282746310005*ser / x));
}

// *** ポアソン分布にしたがった乱数を発生させる関数 ***
// double xm;  // 入力値
// long *idum; // 乱数の初期化
// random number of Poisson distribution for mean of xm
double poidev(double xm, long *idum)
{
	double           em, t, y;
	static double    sq, alxm, g, oldm = (-1.0);

	if (xm < 12.0) {
		if (xm != oldm) {
			oldm = xm;
			g = exp(-xm);
		}
		em = -1.0;
		t = 1.0;
		do {
			++em;
			t *= ran1(idum);
		} while (t > g);
	}
	else {
		if (xm != oldm) {
			oldm = xm;
			sq = sqrt(2.0*xm);
			alxm = log(xm);
			g = (xm*alxm - gammln(xm + 1.0));
		}
		do {
			do {
				y = tan(PI*ran1(idum));
				em = sq*y + xm;
			} while (em < 0.0);
			em = floor(em);
			t = (0.9*(1.0 + y*y)*exp(em*alxm - gammln((em + 1.0)) - g));
		} while (ran1(idum) > t);
	}
	return em;
}
